local ADDON_NAME, Core = ...

--<
local date, time = date, time
local sub, gsub = string.sub, string.gsub
local upper, format, find, match = string.upper, string.format, string.find, string.match
local RGBToColorCode = RGBToColorCode
local UI = Core.UI
local utils = Core.utils
local Sessions = Core.Sessions
-->

local myCharacterName, myBattleTag, myRPAlias, myGUID
local cache = {}


local formatPatterns = {
	generic = "%s%s: %s",
	emoteApostrophe = "%s%s%s",
	emoteSpecial = [[%s%s|TInterface\FriendsFrame\BroadcastIcon:16:16|t %s]],
	EMOTE = "%s* %s %s",

	verbose = {
		WHISPER = "%s%s whispers: %s",
		WHISPER_INFORM = "%sTo %s: %s",
		BN_WHISPER = "%s%s whispers: %s",
		BN_WHISPER_INFORM = "%sTo %s: %s",
		SAY = "%s%s says: %s",
	}
}

setmetatable(formatPatterns, { __index = function(self) return self.generic end })
Core.formatPatterns = formatPatterns


local function chatMsgRestrictedHandler (...)
	local err, target = ...
	local S = Sessions:Get(target)

	if err == "ChatIgnoredHandler" and S then
		S.page:AddMessage(_G.CHAT_RESTRICTED, 1, 0, 0)
	end
end


local function sendMessage (msg, session)
	local queue = table.remove(cache) or {}
	local _max = 255
	local ctype = session.chatType
	local target = session:GetID()
	
	
	if ctype == "EMOTE" and msg:find(myCharacterName) == 1 then
		msg = msg:sub(#myCharacterName+1)
		
	elseif ctype:match("CHANNEL") then
		ctype = ctype:sub(1, 7)	
	end

	if #msg > _max then
		
		while #msg ~= 0 do
			local lastChar = msg:sub(_max, _max)
			
			if lastChar == "" and #msg <= _max then 
				queue[#queue+1] = msg
				msg = ""
				break

			elseif lastChar ~= " " then
				local decrement = _max
				local isWhiteSpace = false
				
				while not isWhiteSpace do
					if msg:sub(decrement, decrement) == " " then
						queue[#queue+1] = msg:sub(0, decrement)
						msg = msg:sub(decrement+1)
						isWhiteSpace = true
						
					elseif decrement == 0 then
						queue[#queue+1] = msg:sub(0, _max)
						msg = ""
						break
						
					else					
						decrement = decrement - 1
					end
					
				end
			
			else
				queue[#queue+1] = msg:sub(0, _max)
				msg = msg:sub(_max+1)
			end
			
		end
	
	elseif msg == "" then
		return UI.editBox:ClearFocus()
	
	else
		queue[1] = msg		
	end

	for i, v in ipairs(queue) do
		queue[i] = v:gsub("^%s", "")
	end	
	
	if type(target) == "string" then
		for i, v in ipairs(queue) do
			ChatThrottleLib:SendChatMessage("NORMAL", "SessionsMSGR", v, ctype, nil, target)
		end

	else
		for i, v in ipairs(queue) do
			BNSendWhisper(target, v)
		end

	end

	session.page:Update(true)
	
	if session.isDM then
		Sessions.lastWhisperTarget = session
	end
	
	table.insert(cache, wipe(queue))
end


local function sendEmote (cmd, msg, session)
	if msg == nil and not UnitName("target") and session then
		msg = session:GetName()
	end
	
	cmd = gsub(cmd, "/", "")
	
	DoEmote(cmd, msg)
end


local function bnetFriendUpdateHandler (...)
	friendIndex = ...
	if friendIndex == nil then 
		return 
	end
	
	payload = C_BattleNet.GetFriendAccountInfo(friendIndex)
	
	local S = Sessions:Get(payload.bnetAccountID)
	if not S then return end
	
	Core.BNetDisplayUpdate(S.tab, 
		payload.gameAccountInfo.characterName,
		payload.gameAccountInfo.clientProgram,
		payload.gameAccountInfo.isOnline
	)
end


local function replyHook ()
	if Sessions.lastWhisperTarget == nil then
		return
	end	

	if not UI:IsShown() then
		UI:Show()
	end

	Sessions:Switch(Sessions.lastWhisperTarget)
	UI:FocusEditBoxFromKeybind()
end


local lastSearch
local function onTellTargetIntercept (target)
	local id = BNet_GetBNetIDAccount(target)
	local name = utils.capitalize(target)

	if id == nil then	
		local fullName = upper(utils.getFullName(name))
		local guid
		
		if UnitName("target") == name and UnitIsPlayer("target") then
			guid = UnitGUID("target")
			
		elseif UnitName("mouseover") == name and UnitIsPlayer("mouseover") then
			guid = UnitGUID("mouseover")			
		else
			local friendInfo = C_FriendList.GetFriendInfo(target)
			if friendInfo then
				guid = friendInfo.guid
			else
				guid = Core.History:GetInfo(fullName)
			end
		end
		
		if not Sessions:Get(fullName) then
			if guid then
			-- Fix for when the client hasn't encountered the GUID yet. Query the GUID, then restart. /shrug
				if GetPlayerInfoByGUID(guid) == nil and lastSearch ~= guid then
					C_Timer.After(0.5, function() onTellTargetIntercept(target) end)
					lastSearch = guid
					return
				end
				
				Sessions.CreateFromGUID(guid)
				lastSearch = nil
			else
				Sessions.CreateFromCharacterName(fullName)
			end
		end
		
		Sessions:Switch(Sessions:Get(fullName))
		
	else
		if not Sessions:Get(id) then
			Sessions.CreateFromID(id)
		end
		
		Sessions:Switch(Sessions:Get(id))
	end
end


local function updateHeaderHook (editBox)
	if editBox:GetAttribute("chatType"):match("WHISPER") == nil then
		Core.G_ChatType = editBox:GetAttribute("chatType")
	end	
end


local function insertItemLink (item)
	if UI.editBox:HasFocus() and IsModifiedClick("CHATLINK") then
		local loc = PaperDollFrame:IsMouseOver() and ItemLocation:CreateFromEquipmentSlot(item:GetID()) or ItemLocation:CreateFromBagAndSlot(item:GetParent():GetID(), item:GetID())
		UI.editBox:Insert(C_Item.GetItemLink(loc))
		
		if StackSplitFrame:IsShown() then
			StackSplitFrame:Hide()
		end
	end
end


local function extractTellTargetHook (editBox, msg)
	local target = msg:match("%s*(.*)")
	
	if ( not target or not target:find("%s") ) then
		return false;
	end
	
	if target:sub(1, 1) == "|" then
		return false
	end
	
	while target:find("%s") do
		target = target:match("(.+)%s+[^%s]*")
		if #GetAutoCompleteResults(target, 1, 0, true, AUTOCOMPLETE_LIST.ALL.include,
			AUTOCOMPLETE_LIST.ALL.exclude) > 0 then 
			break
		end
		
	end

	onTellTargetIntercept(target)
	
	UI:Show()
	editBox:Hide()
	UI.editBox:SetFocus()
	
end


local function sendBNetTellHook (name)
	Core.G_EditBox:Hide()
	UI:Show()
	onTellTargetIntercept(name)
end


local function onPortraitsUpdated ()
	if #Core.orderedTabs == 0 or InCombatLockdown() or not IsInGroup(LE_PARTY_CATEGORY_HOME) then
		return
	end
	
	local numPartyMembers = GetNumGroupMembers(LE_PARTY_CATEGORY_HOME)-1
	
	if numPartyMembers > 0 then		
		for i=1, numPartyMembers do
			local unit = "party".. tostring(i)
			local S = Sessions:Get(UnitGUID(unit))
			
			if S and UnitIsVisible(unit) then
				S:UpdatePortrait(unit)
			end
		end
	end
end


local function GetPattern (ctype, msg)
	if ctype == "EMOTE" and (msg:sub(1, 1) == "|" or msg:match("^%u")) then
		return formatPatterns.emoteSpecial
	elseif ctype == "EMOTE" and msg:sub(1, 1) == "'" then
		return formatPatterns.emoteApostrophe
	else
		return formatPatterns[ctype]
	end
end


local function ParseText (ctype, msg)
	local pattern, hex
	
	if ctype ~= "SAY" then
		pattern = '%b""'; hex = RGBToColorCode(utils.GetRGB("SAY"))
	elseif ctype ~= "EMOTE" then
		pattern = '%b**'; hex = RGBToColorCode(utils.GetRGB("EMOTE"))
	end
	
	msg = gsub(msg, pattern, function(catch)
		local _, pos = msg:find(catch)
		
		if pos and msg:sub(pos+1, pos+2) == "|r" then
			return catch
		else
			return format("%s%s|r", hex, catch)
		end		
	end)
	
	if ctype ~= "EMOTE" then
		msg = gsub(msg, "<?https?://[%w%p]+", function(url)
			if url:match("<") then
				return url
			else
				return format("|H%s|h|cFFC845FA<%s>|r|h", url, url)
			end
		end)
	end
	
	return msg
end


local recentLongMessages = {}

local function onChatMessageReceived (e, ...)
	local ctype = e:gsub("CHAT_MSG_", "")
	
	if ctype == "ADDON" then
		if ... ~= "SessionsMSGR" then
			return
		end
		
		local _, cmd, _, sender = ...
		local S = Sessions:Get(utils.getFullName(sender))
		
		if not S then
			return 
		elseif cmd == "TI01" then
			S:SetTyping()
		end
		
		if not S.isUsingSM then
			S.isUsingSM = true
			UI.editBox:AddOrRemoveOnTextHandler()
		end
		
		return
		
	elseif ctype == "RESTRICTED" then
		return chatMsgRestrictedHandler(...)
	end
	
	local msg, author, language, _, _, _, _, chanIndex, _, _, lineID, GUID, bnetID = ...
	local inbound = (not ctype:match("INFORM$") and true) or false
	local now = date("%I:%M ")
	local S, isBnet

	-- Get or create the appropriate session 
	
	if ctype:match("^WHISPER") then
		S = Sessions:Get(author)
		if not S then S = Sessions.CreateFromGUID(GUID) end
		
	elseif ctype:match("^BN_") then
		S = Sessions:Get(bnetID); isBnet = true
		if not S then S = Sessions.CreateFromID(bnetID) end
	else 
		S = Sessions:Get(ctype)
		if not S then S = Sessions.CreateCustom(unpack(Sessions.preset[ctype])) end
	end

	Sessions:UpdateIfNeeded(author, GUID)
	
	-- First, check if we're appending to the author's previous fontString and exiting
	
	if recentLongMessages[author] then
		if recentLongMessages[author] >= GetTime() then
			for i=1, 6 do
				local info = S.page.historyBuffer:GetEntryAtIndex(i)

				if info and info.extraData and info.extraData[1] == author then
					if info.extraData.n == 1 then
						info.message = ParseText(ctype, info.message.. msg)
						S.page:RefreshLayout()
					else
						info.extraData[2] = ParseText(ctype, info.extraData[2].. msg)
					end
					
					if S.isDM then
						S:AddHistory(inbound, isBnet, now, msg)
					end
					
					break
				end
			end

			return			
		else
			recentLongMessages[author] = nil
		end
	end
	
	if #msg > 240 then
		recentLongMessages[author] = GetTime() + 2
	end
	
	-- Create the appropriate playerLink for this context
	
	local typePreferences = Sessions.preferences[ctype]
	
	local trueinbound = inbound or Core.db.UseVerbose
	local playerLink, displayName, alias
	
	if isBnet then
		playerLink = (inbound or Core.db.UseVerbose) and GetBNPlayerLink(author, author, bnetID) or myBattleTag
		
	elseif trueinbound then
		if typePreferences.useRPAlias then
			alias = Sessions:GetRoleplayAlias(author)
		end
		displayName = Ambiguate(author, "none")
	else
		if typePreferences.useRPAlias then
			alias = myRPAlias
		end
		displayName = myCharacterName
	end
	
	if not playerLink then
		local finalName
		
		if (typePreferences.nameColoured and not alias) or (typePreferences.nameColoured and alias and not alias:match("|r")) then
			local _, englishClass = GetPlayerInfoByGUID(trueinbound and GUID or myGUID)
			local classColour = RAID_CLASS_COLORS[englishClass]
			
			if classColour then
				finalName = classColour:WrapTextInColorCode(alias or displayName)
			end
		else
			finalName = alias or displayName
		end
		
		playerLink = GetPlayerLink(trueinbound and author or myCharacterName, finalName)
	end

	if Core.db.BracketedNames and not ctype:match("EMOTE") then playerLink = format("[%s]", playerLink) end
	
	-- Get the timestamp
	
	local ts
	
	if not typePreferences.tsEnabled then
		ts = ""
	else
		ts = (typePreferences.tsColoured and format("|cFFB3B3B3%s|r", now)) or now
	end
	
	-- Get the appropriate RGB for this context
	
	local r, g, b

	if not inbound and ctype ~= "PARTY_LEADER" then
		r, g, b = .85, .85, .85
	else
		r, g, b = utils.GetRGB(ctype)
	end
	
	-- Put the pieces together!
	
	local body
	
	if ctype == "TEXT_EMOTE" then
		body = format("%s%s", ts, msg:gsub(author, playerLink))
	else
		body = ParseText(ctype, msg)
		body = format(GetPattern(ctype, msg), ts, playerLink, body)
	end

	if S.selectiveModeEnabled and not Sessions:Get(author) and GUID ~= myGUID then
		local notice
		
		if match(msg, myCharacterName) then
			notice = [[|TInterface\OPTIONSFRAME\UI-OptionsFrame-NewFeatureIcon:14:14|t ]]
		end
		
		S.page:AddMessage(
			format("%s%s[by %s: Right-Click To Show]|r", notice or "", RGBToColorCode(.7, .7, .7), author),
			r, g, b, author, body
		)
	else
		if S.isUsingSM then
			S.page:ClearTypingStatus()
		end
		
		S.page:AddMessage(body, r, g, b, author)
	end
	
	-- Lastly, handle notifications and interface things
	
	S.page:Update()
	
	if S.isDM then
		if Core.db.HistoryExpiry > 0 then
			S:AddHistory(inbound, isBnet, now, msg)
		end
		
		if S.isMuted then return end
		
		if UI:IsShown() then
			if not S:IsSelected() then
				S:UpdateNotice(ctype)
				S:StartGlowing()
			end
			
		elseif InCombatLockdown() or Core.db.KeepMinimized then
			if inbound then
				S:UpdateNotice(ctype)
			end
		else
			Sessions:Switch(S)
			UI:Show()
		end
		
		if inbound or Sessions.lastWhisperTarget == nil then
			Sessions.lastWhisperTarget = S
			PlaySound(3081, "Master")
			FlashClientIcon()
		end
		
		if S.tab.tabIndex ~= 1 then
			table.insert(Core.orderedTabs, 1, table.remove(Core.orderedTabs, S.tab.tabIndex))
			Core.Tab_ReanchorAll()
		end
		
		if not isBnet and S.isUsingSM == nil then
			S:Handshake()
		end
		
	elseif not InCombatLockdown() and (not S:IsSelected() or not UI:IsShown()) and not S.isMuted then
		S:UpdateNotice(ctype)
	end
end



-------------------------------------------------------------------
------------------------ Hook dook --------------------------------
-------------------------------------------------------------------



UI.editBox:SetScript("OnTabPressed", function(self)
	if #Core.orderedTabs <= 1 then
		return
	end

	for N, S in pairs(Sessions.active) do
		if S.missedMessages > 0 then
			Sessions:Switch(S)
			return
		end
	end
	
	local current = Sessions.selected.tab.tabIndex
	local nextIndex = Core.orderedTabs[current+1] and current+1 or 1
	
	for N, S in pairs(Sessions.active) do
		if S.tab.tabIndex == nextIndex then
			Sessions:Switch(S)
			break
		end
	end
	
end)


UI.editBox:SetScript("OnSpacePressed", function(this)
	local sample = this:GetText()

	if (sample:sub(1, 1) ~= "/") then
		return
	end
	
	local ctype = sample:sub(2, sample:find("%s")-1):lower()
	local text = gsub(sample, format("/%s ", ctype), "")
	
	if (ctype == "r" or ctype == "reply") and Sessions.lastWhisperTarget ~= nil then
		Sessions:Switch(Sessions.lastWhisperTarget)
		this:SetText(text)

	elseif ctype == "w" or ctype == "whisper" then
		if match(text, "%a") then
			local target = gsub(text, " ", "")
			this:SetText("")
			onTellTargetIntercept(target)
		end
		
	else
		ctype = (utils.any(ctype, "e", "em", "me", "emote")) and "EMOTE"
			or (ctype == "say" or ctype == "s") and "SAY"
			or (ctype == "p" or ctype == "party") and "PARTY"
			or (ctype == "g" or ctype == "guild") and "GUILD"
			
		if not Core.IsTypeEnabled(ctype) then
			return
		end

		local S = Sessions:Get(ctype)
		
		if not S then
			S = Sessions.CreateCustom(unpack(Sessions.preset[ctype]))
		end
		
		if not S:IsSelected() then
			this:SetText("")
			Sessions:Switch(S)
		end
		
		S.chatType = ctype
		
		if ctype == "SAY" or ctype == "EMOTE" then
			S:UpdateTitle()
		end
		
		this:SetText(sample:sub(#ctype+3, #sample))	
	end

	if Sessions.selected then this:SetTextColor(Sessions:GetSelectedRGB()) end
end)


UI.editBox:SetScript("OnEnterPressed", function(self, ...)
	local text = self:GetText()

	if sub(text, 1, 1) == "/" then
		Core.G_EditBox:SetText(text)
		ChatEdit_ParseText(Core.G_EditBox, 1)

	elseif self:HasFocus() and Sessions.selected ~= nil then
		sendMessage(text, Sessions.selected)
	end
	
	UI:SetSliderValues()
	self:SetText("")
	self:ClearFocus()
end)


UI:SetScript("OnEvent", function(self, e, ...)
	if e:match("CHAT_MSG_") then
		onChatMessageReceived(e, ...)
	
	elseif e == "BN_FRIEND_INFO_CHANGED" then
		bnetFriendUpdateHandler(...)

	elseif e == "GROUP_JOINED" then
		local S = Sessions:Get("PARTY") or Sessions.CreateCustom("Party", "PARTY")
		S.page:AddMessage(format(_G.ERR_JOINED_GROUP_S, myCharacterName), 1, 1, 0)
		
	elseif e == "GROUP_LEFT" then
		local S = Sessions:Get("PARTY")
		if S then S.page:AddMessage(_G.ERR_LEFT_GROUP_YOU, 1, 0, 0) end

	elseif e == "MODIFIER_STATE_CHANGED" and UI:IsShown() then
		local key, state = ...
		if key ~= "LCTRL" then return end
		
		if Sessions.selected ~= nil then
			Sessions.selected.page:SetTextCopyable(state > 0 and 1 or nil)
		end
		
		if state == 1 then
			self.resizeBtn:Show()
			
		else
			self:StopMovingOrSizing()
			self.resizeBtn:Hide()
		end
		
	elseif e == "PLAYER_TARGET_CHANGED" then
		if not UnitIsPlayer("target") or not UnitIsVisible("target") then
			return
		end
		
		local S = Sessions:Get(UnitGUID("target"))

		if S then
			S:UpdatePortrait("target")
		end
	
	elseif (e == "GROUP_ROSTER_UPDATE" and not IsInRaid()) or (e == "PORTRAITS_UPDATED") then
		onPortraitsUpdated()
		
	elseif e == "WHO_LIST_UPDATE" then
		Sessions:OnWhoReceived()
		
	elseif e:match("ZONE_CHANGED") and Sessions:Get("NEARBY") then
		Sessions:Get("NEARBY"):UpdateTitle()
	end	
end)


UI:SetScript("OnShow", function(self)
	if not InCombatLockdown() and Core.db.ChatOverrideEnabled then
		SetOverrideBinding(self, true, GetBindingKey("OPENCHAT"), "SMChatOverride")
	end
	
	if Sessions.selected then
		Sessions.selected:MarkAsRead()
	end
	
	for n, session in pairs(Sessions.active) do
		if session.isDM and session.missedMessages > 0 then
			session:StartGlowing()
		end
	end
	
	for i=1, #Core.hub.notices do
		Core.hub.notices[i]:Hide()
	end
end)


UI:SetScript("OnHide", function(self)
	if not InCombatLockdown() then
		SetOverrideBinding(self, true, GetBindingKey("OPENCHAT"), nil)
	end
	
	for i=1, #Core.hub.notices do
		Core.hub.notices[i]:Show()
	end
end)

--[[
UI.leftFrame.inputBox:HookScript("OnEnterPressed", function(self)
	if self:GetText() ~= "" then
		onTellTargetIntercept(self:GetText())
	end
	
	self:SetText("") ; self:ClearFocus()
end)


UI.leftFrame.inputBox:HookScript("OnTextChanged", function(self, byUser)
	if not byUser then
		return
	end
	
	local text = utils.capitalize(self:GetText())
	
	Core.suggestionFrame:ClearList()
	
	if text == "" then 
		return Core.suggestionFrame:Hide() 
	end
	
	Core.suggestionFrame:Show()
	
	local results = GetAutoCompleteResults(text, 4, self:GetCursorPosition(), true, AUTOCOMPLETE_FLAG_ONLINE, AUTOCOMPLETE_FLAG_NONE)

	for i, v in ipairs(results) do
		if Core.suggestionFrame:HasSpace() then
			Core.suggestionFrame:Add(Ambiguate(v.name, "none"), function() onTellTargetIntercept(v.name); self:SetText("") end)
		end
	end
	
end)
]]


hooksecurefunc("ChatEdit_ExtractTellTarget", extractTellTargetHook)
hooksecurefunc("ChatEdit_UpdateHeader", updateHeaderHook)
hooksecurefunc("ChatFrame_ReplyTell", replyHook)
hooksecurefunc("ChatFrame_ReplyTell2", replyHook)
hooksecurefunc("ChatFrame_SendBNetTell", sendBNetTellHook)
hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", insertItemLink)
hooksecurefunc("PaperDollItemSlotButton_OnModifiedClick", insertItemLink)


local function debug ()
	if Core.debugging then
		for i=1, #Core.debugging do
			local err = Core.debugging[i]
			utils.SMPrint(err)
		end
	else
		utils.SMPrint("Nothing to report!")
	end
end



SlashCmdList["SESSIONS_TALKTO"] = function(msg)
	onTellTargetIntercept(msg)
	UI:Show()
end

SLASH_SESSIONS_TALKTO1 = "/talkto"
SLASH_SESSIONS_TALKTO2 = "/tt"


SlashCmdList["SESSIONS_DEBUG"] = debug

SLASH_SESSIONS_DEBUG1 = "/smdebug"

function Core.UpdatePlayerInfo ()
	myCharacterName = UnitName("Player")
	myGUID = UnitGUID("Player")
	myRPAlias = Sessions:GetRoleplayAlias(myCharacterName)
	myBattleTag = BNConnected() and BNet_GetTruncatedBattleTag(select(2, BNGetInfo())) or "Me"
	Core.myCharacterName = myCharacterName
	Core.myRealmName = gsub(GetRealmName(), " ", "")
end

