local ADDON_NAME, Core = ...

--<
local UI = Core.UI
local utils = Core.utils
local Sessions = Core.Sessions
local CreateFrame = CreateFrame
-->


local ctypePrefs = {
	tsColoured=true,
	tsEnabled=true,
	nameColoured=false,
	useRPAlias=false,
}


local function new (tbl)
	local newTable = {}
	for x, y in pairs(tbl) do
		newTable[x] = y
	end
	return newTable
end


local cfg = {
	Version = false,
	LastHistoryCleanUp = false,
	BGOpacity = .70,
	UseVerbose = false,
	BracketedNames = true,
	StripColours = true,
	UILayout = {},
	HubLayout = {},
	HistoryExpiry = 14,
	CloseOnESC = true,
	KeepMinimized = false,
	ChatOverrideEnabled = true,
	PartyEnabled = false,
	GuildEnabled = false,
	NearbyChatEnabled = false,
	TextSize = 14,
	TextFont = "",
	TextOutline = false,
	ChatPreferences = {
		GUILD = new(ctypePrefs),
		PARTY = new(ctypePrefs),
		WHISPER = new(ctypePrefs),
		EMOTE = new(ctypePrefs),
		SAY = new(ctypePrefs),
	},
}


Core.defaultConfig = cfg


local Main, Chat, TF
local contextMenu
local whatIsNew


-------------------------------------------------------------------
-- QuickOptions context menu

local function setExpiry (info)
	SESMSG_Profile["HistoryExpiry"] = info.value
end


local function checkExpiry (info)
	return SESMSG_Profile["HistoryExpiry"] == info.value
end


local function setOpacity (info)
	UI.BG:SetColorTexture(0, 0, 0, info.value)
	SESMSG_Profile["BGOpacity"] = info.value
end


local function checkOpacity (info)
	return info.value == SESMSG_Profile["BGOpacity"]
end


local function nearbyFilter (self, event, ...)
	if Core.db.NearbyChatEnabled then
		return true
	else return false, ...
	end
end
local function guildFilter (...)
	if Core.db.GuildEnabled then
		return true
	else return false, ...
	end
end


local function toggle_PartyChat (byUser)
	if Core.db.PartyEnabled then
		FrameUtil.RegisterFrameForEvents(UI, Sessions.partyEvents)
		
		if byUser and IsInGroup(LE_PARTY_CATEGORY_HOME) then
			Sessions.CreateCustom("Party", "PARTY")
		end
	else
		FrameUtil.UnregisterFrameForEvents(UI, Sessions.partyEvents)
		Sessions:CloseIfActive("PARTY")
	end
end


local function toggle_GuildChat (byUser)
	if SESMSG_Profile["GuildEnabled"] and IsInGuild() then
		UI:RegisterEvent("CHAT_MSG_GUILD")
		ChatFrame_AddMessageEventFilter("CHAT_MSG_GUILD", guildFilter)
		if byUser then
			utils.SMPrint("Now taking over for Guild messages")
			Sessions.CreateCustom("Guild", "GUILD")
		end
	else
		UI:UnregisterEvent("CHAT_MSG_GUILD")
		ChatFrame_RemoveMessageEventFilter("CHAT_MSG_GUILD", guildFilter)
		Sessions:CloseIfActive("GUILD")
	end
end


local function toggle_NearbyChat (byUser)
	if Core.db.NearbyChatEnabled then
		FrameUtil.RegisterFrameForEvents(UI, Sessions.nearbyEvents)

		ChatFrame_AddMessageEventFilter("CHAT_MSG_EMOTE", nearbyFilter)
		ChatFrame_AddMessageEventFilter("CHAT_MSG_TEXT_EMOTE", nearbyFilter)
		ChatFrame_AddMessageEventFilter("CHAT_MSG_SAY", nearbyFilter)
		
		if byUser then
			Sessions.CreateCustom("Nearby", "EMOTE", "SAY")
			utils.SMPrint("Now taking over for Say and Emote messages")
		end
	else
		FrameUtil.UnregisterFrameForEvents(UI, Sessions.nearbyEvents)

		ChatFrame_RemoveMessageEventFilter("CHAT_MSG_EMOTE", nearbyFilter)
		ChatFrame_RemoveMessageEventFilter("CHAT_MSG_TEXT_EMOTE", nearbyFilter)
		ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SAY", nearbyFilter)
		
		Sessions:CloseIfActive("Nearby")
	end
end


local function toggle_BlizzardKeyBindingFrame ()
	if not KeyBindingFrame then
		KeyBindingFrame_LoadUI()
	end
	
	for i, cat in ipairs(KeyBindingFrame.categoryList.buttons) do
	
		if cat:GetText() == "AddOns" then
			KeyBindingFrame:Show()
			KeybindingsCategoryListButton_OnClick(cat)
		end
		
	end	
end


local function toggle_BlizzardOptionsFrame ()
	InterfaceOptionsFrame_Show()
	InterfaceOptionsFrame_OpenToCategory(Chat.frame)
end


local function toggle_ChatOverride ()
	SetOverrideBinding(UI, true, GetBindingKey("OPENCHAT"), Core.db.ChatOverrideEnabled and "SMChatOverride" or nil)
end


local function toggle_CloseOnESCPressed ()
	if SESMSG_Profile["CloseOnESC"] then
		table.insert(UISpecialFrames, UI:GetName())

	else
		for i, frameName in ipairs(UISpecialFrames) do
			if frameName == UI:GetName() then
				table.remove(UISpecialFrames, i)
			end
		end
	end
end


local execute = {
	["CloseOnESC"] = toggle_CloseOnESCPressed,
	["ChatOverrideEnabled"] = toggle_ChatOverride,
	["NearbyChatEnabled"] = toggle_NearbyChat,
	["PartyEnabled"] = toggle_PartyChat,
	["GuildEnabled"] = toggle_GuildChat,
}

local function context_changeSetting (info)
	local setting = info.value
	local func = execute[setting]
	Core.db[setting] = not Core.db[setting]
	
	if func then func(true) end
end

local function context_isChecked (info)
	return Core.db[info.value]
end

-------------------------------------------------------------------

local function createContextMenu ()
	local titleGradient = "|cFF417dc1Sess|r|cFF2effebions|r |cFFf28500Mess|r|cFFeb2effenger|r"
	contextMenu = {
		{text=titleGradient.." v".. Core.VERSION, isTitle=true, notCheckable=true},
		{text="|cFF417dc1What's New?|r", notCheckable=true, keepShownOnClick=true, func=function() utils.SMPrint(whatIsNew) end},
		{text="|cFFf28500Open Options Panel|r", notCheckable=true, func=toggle_BlizzardOptionsFrame},
		{text=[[|TInterface\COMMON\UI-TooltipDivider:8:144|t]], notClickable=true, notCheckable=true},
		{text="Core", isTitle=true, notCheckable=true},
		{
			text="Background Opacity",
			hasArrow=true,
			notCheckable=true,
			keepShownOnClick=true,
			menuList={
				{text="30 %", value=.30, checked=checkOpacity, func=setOpacity},
				{text="70 %", value=.70, checked=checkOpacity, func=setOpacity},
				{text="85 %", value=.85, checked=checkOpacity, func=setOpacity},
				{text="100 %", value=1, checked=checkOpacity, func=setOpacity},
			},
		},
		{
			text="Close UI On ESC",
			value="CloseOnESC",
			checked=context_isChecked,
			func=context_changeSetting,
			keepShownOnClick=true,
			isNotRadio=true,
		},
		{
			text="Keep Minimized",
			value="KeepMinimized",
			checked=context_isChecked,
			func=context_changeSetting,
			keepShownOnClick=true,
			isNotRadio=true,
			tooltipTitle="Keep Minimized",
			tooltipText="Prevents the main UI from popping up when receiving whispers",
			tooltipOnButton=true,
		},
		{
			text="Enable Chat Override",
			value="ChatOverrideEnabled",
			checked=context_isChecked,
			func=context_changeSetting,
			keepShownOnClick=true,
			isNotRadio=true,
			tooltipTitle="OpenChat Override",
			tooltipText="Pressing the OpenChat key (default <ENTER>) will instead focus Sessions when the UI is shown.",
			tooltipOnButton=true,
			
		},		
		{
			text="Toggle Resize Button",
			func=function() UI.resizeBtn:SetShown(not UI.resizeBtn:IsShown()) end,
			notCheckable=true,
			keepShownOnClick=true,
			tooltipTitle="Tip!",
			tooltipText="Hold down CTRL to show resize button",
			tooltipOnButton=true,
		},
		{
			text="Change Chat Colours",
			notCheckable=true,
			func=function() ChatConfigFrame:Show() end,			
		},
		{
			text="Set Keybinds",
			notCheckable=true,
			func=toggle_BlizzardKeyBindingFrame,
		},
		{text=[[|TInterface\COMMON\UI-TooltipDivider:8:144|t]], notClickable=true, notCheckable=true},
		{text="Channels", isTitle=true, notCheckable=true},
		{
			text="Enable Party Chat",
			value="PartyEnabled",
			keepShownOnClick=true,
			checked=context_isChecked,
			isNotRadio=true,
			func=context_changeSetting
		},
		{
			text="Enable Guild Chat",
			value="GuildEnabled",
			keepShownOnClick=true, 
			checked=context_isChecked,
			isNotRadio=true,
			func=context_changeSetting
		},
		{
			text="Enable Nearby Chat",
			value="NearbyChatEnabled",
			keepShownOnClick=true,
			tooltipTitle="Nearby Chat",
			tooltipText="Say & Emote",
			tooltipOnButton=true,
			checked=context_isChecked,
			isNotRadio=true,
			func=context_changeSetting
			
		},
		{text=[[|TInterface\COMMON\UI-TooltipDivider:8:144|t]], notClickable=true, notCheckable=true},
		{text="History", isTitle=true, notCheckable=true},
		{
			text="Open History Viewer",
			notCheckable=true,
			func=Core.History.Open,
		},
		{
			text= "History Expiry",
			hasArrow=true,
			notCheckable=true,
			keepShownOnClick=true,
			menuList={
				{text="Disabled", value=0, checked=checkExpiry, func=setExpiry},
				{text="14 Days", value=14, checked=checkExpiry, func=setExpiry},
				{text="1 Month", value=30, checked=checkExpiry, func=setExpiry},
				{text="2 Months", value=60, checked=checkExpiry, func=setExpiry},
				{text="4 Months", value=120, checked=checkExpiry, func=setExpiry},
			}
		},
		{text="Cancel", notCheckable=true},
	}
	
end


function Core.ShowOptions ()
	EasyMenu(contextMenu, UI.floatingMenu, "cursor", 0, 110, "MENU")
end


function Core.CreatePublicSessions ()
	if IsInGroup(LE_PARTY_CATEGORY_HOME) and Core.db.PartyEnabled and not Sessions:Get("PARTY") then
		toggle_PartyChat()
		Sessions.CreateCustom("Party", "PARTY")
	end

	if IsInGuild() and Core.db.GuildEnabled and not Sessions:Get("GUILD") then
		toggle_GuildChat()
		Sessions.CreateCustom("Guild", "GUILD")
	end
	
	if Core.db.NearbyChatEnabled and not Sessions:Get("Nearby") then
		toggle_NearbyChat()
		Sessions.CreateCustom("Nearby", "EMOTE", "SAY")
	end
end


function Core.IsTypeEnabled (ctype)
	if ctype == "SAY" or ctype == "EMOTE" or ctype == "TEXT_EMOTE" then
		return Core.db.NearbyChatEnabled
		
	elseif ctype == "PARTY" or ctype == "PARTY_LEADER" then
		return Core.db.PartyEnabled
		
	elseif ctype == "GUILD" then
		return Core.db.GuildEnabled
	else
		return false
	end
end


local function setFramePosition (frame, setting)
	local layout = Core.db[setting]
	frame:ClearAllPoints()
	
	if not next(layout) then
		frame:SetPoint("CENTER")
		utils.SaveFramePosition(frame, setting)
		return
	end
	
	frame:SetPoint(layout.point, UIParent, layout.rel, layout.x, layout.y)
end


local function setFrameSize (frame, setting)
	local layout = Core.db[setting]
	
	if not next(layout) or not layout.width then
		frame:SetSize(500, 300)
		layout.width, layout.height = frame:GetSize()
		return 
	end
	
	frame:SetSize(layout.width, layout.height)
end


function Core.LoadPreferences ()
	UI.BG:SetColorTexture(0, 0, 0, Core.db.BGOpacity)
	
	setFramePosition(UI, "UILayout")
	setFramePosition(Core.hub, "HubLayout")
	Core.hub:SetAnchor()
	setFrameSize(UI, "UILayout")
	UI.editBox:SetWidth(UI.lowerFrame:GetWidth())
	
	toggle_NearbyChat()
	toggle_PartyChat()
	toggle_CloseOnESCPressed()
	
	createContextMenu()
	Sessions:SetPreferences()
end


-------------------------------------------------------------------
-- Options Panel



local isDrawn = false
local onOkayTimer = 0
local fallback = {}


local desc = {
	["tsEnabled"] = "Enable Timestamps",
	["tsColoured"] = "Grey Timestamps",
	["nameColoured"] = "Colourize Names By Class",
	["useRPAlias"] = "Enable Roleplay Names (Requires XRP, TotalRP3, or MyRolePlay)",
	["ApplyToAll"] = "* Apply Settings to All Chat Types",
	["BracketedNames"] = "Enable Bracketed [Names]",
	["UseVerbose"] = "Enable Verbose Prefixes for Say and Whispers",
	["StripColours"] = "Remove Custom Colours from Roleplay Names",
	["TextOutline"] = "Text Outline",
}


local function dropdown_IsChecked (info)
	return info.value == info.owner.current
end

local function checkbutton_OnCheck (self)
	Core.db[self.setting] = self:GetChecked()
end

local function checkbutton_SetEnabled (self, enable)
	if enable then
		self:Enable()
		self.label:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	else
		self:Disable()
		self.label:SetTextColor(GRAY_FONT_COLOR.r, GRAY_FONT_COLOR.g, GRAY_FONT_COLOR.b)
	end
end

local function applyToAll ()
	for ctype in pairs(cfg.ChatPreferences) do
		for pref in pairs(ctypePrefs) do
			Core.db.ChatPreferences[ctype][pref] = Chat[pref]:GetChecked()
		end
	end
end


local Panel = {height=10, indent=15, xSpacing=15, ySpacing=10}
Panel.__index = Panel


function Panel:New (name, parent)
	local tbl = {}
	setmetatable(tbl, Panel)

	tbl.frame = CreateFrame("Frame", nil, parent or UIParent)
	
	tbl.frame.okay = self.OnOkay
	tbl.frame.cancel = self.OnCancel
	tbl.frame.name = name
	tbl.frame.parent = parent and parent.name
	
	InterfaceOptions_AddCategory(tbl.frame)
	tbl:SetTitle()
	return tbl
end


function Panel.OnCancel ()
	for i, v in pairs(fallback) do
		Core.db[i] = v
	end
end


function Panel.OnOkay ()
	-- This gets called 3 times in a row by InterfaceOptions for some reason and I don't like it
	if onOkayTimer > time() then return end
	onOkayTimer = time() + 2

	if Chat.ApplyToAll:GetChecked() then
		applyToAll()
	end
	
	Sessions:SetPreferences()
end


function Panel:Attach (item, inline, level, yOffset)
	if inline then
		item:SetPoint("LEFT", self.anchor, "RIGHT", self.xSpacing, 0)
	else
		yOffset = yOffset or 0
		item:SetPoint("TOPLEFT", self.indent*(level or 1), -(self.height + yOffset))
		self.height = self.height + (item:GetHeight() + self.ySpacing + yOffset)
	end
	
	self.anchor = item
end


function Panel:CreateString (text, fObject)
	local str = self.frame:CreateFontString(nil)
	str:SetFontObject(fObject or "GameFontWhite")
	str:SetJustifyH("LEFT")
	str:SetText(text)
	
	return str
end


function Panel:SetTitle ()
	self.title = self:CreateString(self.frame.name, "GameFontNormalLarge")
	self:Attach(self.title)
end


function Panel:AddSection (text)
	local section = self:CreateString(text)
	self:Attach(section, false, 2, (self.height < 40 and 0 or 20))
end


function Panel:CreateSlider (setting, MIN, MAX, func)
	local name = "SMOptionsPanelSlider_"..setting
	local f = CreateFrame("Slider", name, self.frame, "OptionsSliderTemplate")
	f:SetScript("OnValueChanged", func)
	
	f.value = self:CreateString()
	f.value:SetPoint("CENTER", f, "CENTER", 0, -12)
	
	f:SetMinMaxValues(MIN, MAX)
	f:SetValue(Core.db[setting])
	f:SetValueStep(1); f:SetObeyStepOnDrag(true)
	
	_G[name.. "Low"]:SetText(MIN)
	_G[name.. "High"]:SetText(MAX)
	self:Attach(f, false, 3, 5)
	self[setting.. "Slider"] = f
end


function Panel:CreateDropDown (setting, label, list)
	local f = CreateFrame("Frame", "SMOptionsPanelDD_".. setting, self.frame, "UIDropDownMenuTemplate")
	UIDropDownMenu_SetWidth(f, 150)
	UIDropDownMenu_SetText(f, label)
	
	UIDropDownMenu_Initialize(f, function(this)
		local info = UIDropDownMenu_CreateInfo()
		
		for n, v in pairs(list) do
			info.text=utils.capitalize(n); info.value=n; info.owner=f;
			info.checked=dropdown_IsChecked
			info.func=this.OnClick
			UIDropDownMenu_AddButton(info)
		end
	end)

	f.Button:SetScript("OnClick", function(_,_, down) 
		if not down then return end
		ToggleDropDownMenu(1, nil, f)
	end)
	
	self:Attach(f, false, 2)
	self[setting.. "_DropDown"] = f
end


function Panel:CreateCheckButton (setting, inline, level, func)
	local f = CreateFrame("CheckButton", nil, self.frame, "SessionsCheckButtonTemplate")
	f:SetScript("OnClick", func)
	f:SetChecked(Core.db[setting])
	f.setting = setting
	f.label:SetText(desc[setting])
	self:Attach(f, inline, level)
	self[setting] = f
	return f
end

--[[ Not yet needed anywhere
function Panel:CreateRadioGroup (setting, ...)
	self[setting.. "RadioGroup"] = {}
	
	for i=1, select("#", ...) do
		local value = select(i, ...)
		local f = CreateFrame("CheckButton", "SMOptionsPanelRB_".. value, self.frame, "UIRadioButtonTemplate")
		f:SetChecked(Core.db[setting] == value)
		f.value = value
		f.setting = setting
		
		table.insert(self[setting.. "RadioGroup"], f)
		_G[f:GetName().. "Text"]:SetText(value)
		self:Attach(f, false, 3)
		
		f:SetScript("OnClick", function(this)
			local groupName = this.setting.. "RadioGroup"
			
			for i=1, #self[groupName] do
				self[groupName][i]:SetChecked(false)
			end
			
			this:SetChecked(true)
			Core.db[this.setting] = this.value
		end)
	end
end
--]]

local function panel_Draw ()
	Main:AddSection("(under construction)")
	
	TF:AddSection("Change Sessions' chat font")
	local scroller = CreateFrame("ScrollFrame", nil, TF.frame, "SessionsQuickListTemplate")
	scroller.items = {}
	TF:Attach(scroller, false, 3)
	
	local function button_OnClick (self)
		for i=1, #scroller.items do
			scroller.items[i]:UnlockHighlight()
		end
		self:LockHighlight()
		Core.db.TextFont = self.label:GetText()
	end
	local index = 0
	for name, path in pairs(Core.LSM:HashTable("font")) do
		local fontObject = CreateFont("SM_Demo16_".. name)
		fontObject:SetFont(path, 16)
		
		if fontObject:GetFont() then
			local b = CreateFrame("Button", nil, scroller.ListFrame, "SessionsQuickItemTemplate")
			scroller.items[#scroller.items+1] = b
			b:SetPoint("TOPLEFT", scroller.ListFrame, "TOPLEFT", 2, -index * b:GetHeight())
			b.label:SetFontObject(fontObject)
			b.label:SetText(name)
			b:SetScript("OnClick", button_OnClick)
			if Core.ChatFont:GetFont() == path then b:LockHighlight() end
			index = index+1
		else
			utils.AddError(string.format("Found an oddity with font %s (%s): Font skipped.", name, path))
		end
	end

	TF:AddSection("Configure text size")
	
	TF:CreateSlider("TextSize", 10, 18, function(this, newValue)
		if Core.db.TextSize ~= newValue or not this.value:GetText() then
			this.value:SetText(string.format("(%s)", newValue)); Core.db.TextSize = newValue
		end
	end)
	
	TF:AddSection("Enable or disable text outline")
	
	TF:CreateCheckButton("TextOutline", false, 3, function(this) Core.db.TextOutline = this:GetChecked() and "OUTLINE" or false end)

	Chat:AddSection("Configure how names are displayed")
	
	Chat:CreateCheckButton("UseVerbose", false, 3, checkbutton_OnCheck)
	
	Chat:CreateCheckButton("BracketedNames", false, 3, checkbutton_OnCheck)
	
	Chat:CreateCheckButton("StripColours", false, 3, checkbutton_OnCheck)

	Chat:AddSection("Configure chat preferences for each chat type, or use the same preferences for all")
	
	Chat:CreateDropDown("ChatPreferences", "Select chat type", Core.db.ChatPreferences)
	local dropdown = Chat.ChatPreferences_DropDown
	
	dropdown.OnClick = function(btn)
		btn.owner.current = btn.value;
		UIDropDownMenu_SetText(btn.owner, btn.value)

		for setting, bool in pairs(Core.db.ChatPreferences[btn.value]) do
			if not Chat[setting]:IsEnabled() then
				checkbutton_SetEnabled(Chat[setting], true)
			end
			Chat[setting]:SetChecked(bool)
		end
	end
	
	Chat:CreateCheckButton("ApplyToAll", true, nil, function(this)
		local checked = this:GetChecked()
		_G["UIDropDownMenu_".. (checked and "Disable" or "Enable").. "DropDown"](dropdown)
		
		for setting, bool in pairs(ctypePrefs) do
			if checked and not Chat[setting]:IsEnabled() then
				checkbutton_SetEnabled(Chat[setting], true)
				
			elseif not checked and not dropdown.current then
				checkbutton_SetEnabled(Chat[setting], false)
			end
		end
	end)
	
	local callback = function(this)
		if Chat.ApplyToAll:GetChecked() then
			applyToAll()
		else
			Core.db.ChatPreferences[dropdown.current][this.setting] = this:GetChecked()
		end
	end
	
	for setting in pairs(ctypePrefs) do
		local cb = Chat:CreateCheckButton(setting, false, 4, callback)
		checkbutton_SetEnabled(cb, false)
	end
	
	Chat.frame:SetScript("OnShow", function()
		if dropdown.current then
			for setting, bool in pairs(Core.db.ChatPreferences[dropdown.current]) do
				Chat[setting]:SetChecked(bool)
			end
		end

	end)
	
	TF.TextSizeSlider:SetValue(Core.db.TextSize)
	
	isDrawn = true
end


local function panel_OnShow ()
	for i, v in pairs(Core.db) do
		fallback[i] = v
	end
end


Main = Panel:New("Sessions Messenger")
Chat = Panel:New("Chat Preferences", Main.frame)
TF = Panel:New("Text & Font", Main.frame)

-- Only creating all this junk if-when it's needed
InterfaceOptionsFrame:HookScript("OnShow", function()
	if not isDrawn then panel_Draw() end panel_OnShow()
end)

Core.openBlizzardOptionsFrame = toggle_BlizzardOptionsFrame


whatIsNew =
([[Thank you for using Sessions! Here's what's new in v%s:

]]):format(Core.VERSION)


